

#include <EEPROM.h>     // We are going to read and write PICC's UIDs from/to EEPROM
#include <SPI.h>        // RC522 Module uses SPI protocol
#include <MFRC522.h>  // Library for Mifare RC522 Devices
#include <LiquidCrystal_I2C.h> // Interface I2C <=> LCD

#define relay 4     // Set Relay Pin
#define wipeB 3     // Button pin for WipeMode

bool programMode = false;  // initialize programming mode to false

uint8_t successRead;    // Variable integer to keep if we have Successful Read from Reader

byte storedCard[4];   // Stores an ID read from EEPROM
byte readCard[4];   // Stores scanned ID read from RFID Module
byte masterCard[4];   // Stores master card's ID read from EEPROM

// Create MFRC522 instance.
#define SS_PIN 10
#define RST_PIN 9
MFRC522 mfrc522(SS_PIN, RST_PIN);
LiquidCrystal_I2C lcd(0x27,20,4);  // set the LCD address to 0x27 for a 20 chars and 4 lines display

///////////////////////////////////////// Setup ///////////////////////////////////
void setup() {
  //Arduino Pin Configuration
  lcd.init();                      // initialize the lcd 
  lcd.backlight();
  pinMode(wipeB, INPUT_PULLUP);   // Enable pin's pull up resistor
  pinMode(relay, OUTPUT);
  digitalWrite(relay, LOW);    // Make sure door is locked

  //Protocol Configuration
  SPI.begin();           // MFRC522 Hardware uses SPI protocol
  mfrc522.PCD_Init();    // Initialize MFRC522 Hardware

  //If you set Antenna Gain to Max it will increase reading distance
  //mfrc522.PCD_SetAntennaGain(mfrc522.RxGain_max);

  lcd.print("  Control d'acces:");   // For debugging purposes
  lcd.setCursor(1, 1);
  lcd.print("Feelgood & Nanard");
  delay (2000);
  ShowReaderDetails();  // Show details of PCD - MFRC522 Card Reader details
  

  //Wipe Code - If the Button (wipeB) Pressed while setup run (powered on) it wipes EEPROM
  if (digitalRead(wipeB) == LOW) {  // when button pressed pin should get low, button connected to ground
    lcd.clear();
    lcd.print("Bouton Presse ==>");
    lcd.setCursor(0, 1);
    lcd.print("Si 10s  'Eff. Total' ");
    bool buttonState = monitorWipeButton(10000); // Give user enough time to cancel operation
    if (buttonState == true && digitalRead(wipeB) == LOW) {    // If button still be pressed, wipe EEPROM
      lcd.setCursor(0, 2);
      lcd.print("Eff. en Cours...");
      for (uint16_t x = 0; x < EEPROM.length(); x = x + 1) {    //Loop end of EEPROM address
        if (EEPROM.read(x) == 0) {              //If EEPROM address 0
          // do nothing, already clear, go to the next address in order to save time and reduce writes to EEPROM
        }
        else {
          EEPROM.write(x, 0);       // if not write 0 to clear, it takes 3.3mS
        }
      }
      lcd.setCursor(1, 3);
      lcd.print("Totalement Effacee !");
      delay(4000);

    }
    else {
      lcd.setCursor(0, 3);
      lcd.print("Effacement Arrete!"); // Show some feedback that the wipe button did not pressed for 15 seconds
      delay(4000);
    }
  }
  // Check if master card defined, if not let user choose a master card
  // This also useful to just redefine the Master Card
  // You can keep other EEPROM records just write other than 143 to EEPROM address 1
  // EEPROM address 1 should hold magical number which is '143'
  if (EEPROM.read(1) != 143) {
    lcd.clear();
    lcd.print("Pas de MasterCard");
    lcd.setCursor(0, 1);
    lcd.print("Scanner pour Definir");
    do {
      successRead = getID();            // sets successRead to 1 when we get read from reader otherwise 0
      delay(200);
    }
    while (!successRead);                  // Program will not go further while you not get a successful read
    for ( uint8_t j = 0; j < 4; j++ ) {        // Loop 4 times
      EEPROM.write( 2 + j, readCard[j] );  // Write scanned PICC's UID to EEPROM, start from address 3
    }
    EEPROM.write(1, 143);                  // Write to EEPROM we defined Master Card.
    lcd.setCursor(0, 3);
    lcd.print("MasterCard Definie");
    delay (2000);
  }

  lcd.clear();
  lcd.print(" UID MasterCard :");
  lcd.setCursor(5, 1);
  for ( uint8_t i = 0; i < 4; i++ ) {          // Read Master Card's UID from EEPROM
    masterCard[i] = EEPROM.read(2 + i);    // Write it to masterCard    
    lcd.print(masterCard[i], HEX);
      }
  delay (5000);
  lcd.setCursor(0, 2);
  lcd.print("Pret a Scanner");
  lcd.setCursor(2, 3);
  lcd.print("le prochain Badge");

}


///////////////////////////////////////// Main Loop ///////////////////////////////////
void loop () {
  do {
    successRead = getID();  // sets successRead to 1 when we get read from reader otherwise 0
    // When device is in use if wipe button pressed for 10 seconds initialize Master Card wiping
    if (digitalRead(wipeB) == LOW) { // Check if button is pressed
      lcd.clear();
      lcd.print("Attention MasterCard");
      lcd.setCursor(0, 1);
      lcd.print("Effacee dans 10 sec.");
      bool buttonState = monitorWipeButton(10000); // Give user enough time to cancel operation
      if (buttonState == true && digitalRead(wipeB) == LOW) {    // If button still be pressed, wipe EEPROM
        EEPROM.write(1, 0);                  // Reset Magic Number.
        lcd.setCursor(1, 2);
        lcd.print("MasterCard Effacee");
        lcd.setCursor(0, 3);
        lcd.print("Reset puis re-ecrire");
        while (1);
      }
      lcd.clear();
      lcd.print("Faute Eff MasterCard");
    }
  }
  while (!successRead);   //the program will not go further while you are not getting a successful read
  if (programMode) {
    if ( isMaster(readCard) ) { //When in program mode check First If master card scanned again to exit program mode
      lcd.clear();
      lcd.print("La MasterCard a ete");
      lcd.setCursor(0, 1);
      lcd.print("scannee, vous etes");
      lcd.setCursor(0, 2);
      lcd.print("resorti du mode");
      lcd.setCursor(0, 3);
      lcd.print(" 'Programmation'");
      programMode = false;
      return;
    }
    else {
      if ( findID(readCard) ) { // If scanned card is known delete it
        lcd.clear();
        lcd.setCursor(0, 2);
        lcd.print("Ce Badge est connu");
        lcd.setCursor(0, 3);
        lcd.print(" ==>Effacement !!!");
        delay(4000);
        deleteID(readCard);
        lcd.clear();
        lcd.print("Scanner Badge pour");
        lcd.setCursor(0, 1);
        lcd.print("Ajouter ou Effacer");
        lcd.setCursor(0, 2);
        lcd.print("ou Master Card pour");
        lcd.setCursor(0, 3);
        lcd.print("sortie mode Program");
      }
      else {                    // If scanned card is not known add it
        lcd.clear();
        lcd.setCursor(0, 2);
        lcd.print("Ce Badge inconnu, ");
        lcd.setCursor(0, 3);
        lcd.print("==> Ajout en EEPROM");
        delay(4000);
        writeID(readCard);
        lcd.clear();
        lcd.print("Scanner Badge pour");
        lcd.setCursor(0, 1);
        lcd.print("Ajouter ou Effacer");
        lcd.setCursor(0, 2);
        lcd.print("ou Master Card pour");
        lcd.setCursor(0, 3);
        lcd.print("sortie mode Program");
        
      }
    }
  }
  else {
    if ( isMaster(readCard)) {    // If scanned card's ID matches Master Card's ID - enter program mode
      programMode = true;
      lcd.clear();
      lcd.print("MasterCard Detectee");
      lcd.setCursor(0, 1);
      lcd.print("Entree Mode Program:");
      uint8_t count = EEPROM.read(0);   // Read the first Byte of EEPROM that
      lcd.setCursor(0, 2);
      lcd.print("J'ai ");     // stores the number of ID's in EEPROM
      lcd.print(count);
      lcd.print(" Badge(s)");
      lcd.setCursor(3, 3);
      lcd.print("en Memoire !");
      delay(5000);
      lcd.clear();
      lcd.print("Approcher un Badge");
      lcd.setCursor(0, 1);
      lcd.print("pour Ajout ou Suppr");
      lcd.setCursor(0, 2);
      lcd.print("Ou Master Card");
      lcd.setCursor(0, 3);
      lcd.print("pour sortie Program");
    }
    else {
      if ( findID(readCard) ) { // If not, see if the card is in the EEPROM
        granted();         // Open the door lock for 300 ms
      }
      else {      // If not, show that the ID was not valid
        denied();
      }
    }
  }
}

/////////////////////////////////////////  Access Granted    ///////////////////////////////////
void granted () {
        lcd.clear();
        lcd.print("     Dac O Dac...");
        lcd.setCursor(2, 2);
        lcd.print("Acces Autorise !");
  digitalWrite(relay, HIGH);     // Unlock door!
  delay(3000);          // Hold door lock open for given seconds
  digitalWrite(relay, LOW);    // Relock door
  delay(7000);            // Hold green LED on for a second
    lcd.noBacklight();
}

///////////////////////////////////////// Access Denied  ///////////////////////////////////
void denied() {
        lcd.clear();
        lcd.print("Vous n'avez pas");
        lcd.setCursor(3, 1);
        lcd.print("l'acces! S.V.P.");
        lcd.setCursor(0, 2);
        lcd.print("Utilisez un Badge");
        lcd.setCursor(0, 3);
        lcd.print("enregistre et connu");
  delay(7000);
    lcd.noBacklight();
}


///////////////////////////////////////// Get PICC's UID ///////////////////////////////////
uint8_t getID() {
  // Getting ready for Reading PICCs
  if ( ! mfrc522.PICC_IsNewCardPresent()) { //If a new PICC placed to RFID reader continue
    return 0;
  }
  if ( ! mfrc522.PICC_ReadCardSerial()) {   //Since a PICC placed get Serial and continue
    return 0;
  }
  // There are Mifare PICCs which have 4 byte or 7 byte UID care if you use 7 byte PICC
  // I think we should assume every PICC as they have 4 byte UID
  // Until we support 7 byte PICCs
  lcd.backlight();
  for ( uint8_t i = 0; i < 4; i++) {  //
    readCard[i] = mfrc522.uid.uidByte[i];
  }
  mfrc522.PICC_HaltA(); // Stop reading
  return 1;
}

void ShowReaderDetails() {
  // Get the MFRC522 software version
  byte v = mfrc522.PCD_ReadRegister(mfrc522.VersionReg);
  lcd.setCursor(0, 2);
  lcd.print("MFRC522 Soft Ver.:");
  lcd.setCursor(5, 3);
  lcd.print(v, HEX);
  if (v == 0x91){
    lcd.print(" = v1.0");
    delay(4000);
  }
  else if (v == 0x92){
    lcd.print(" = v2.0");
    delay(4000);
  }
  else{
    lcd.print("Non reconnu !");
    delay(4000);
    }
  // When 0x00 or 0xFF is returned, communication probably failed
  if ((v == 0x00) || (v == 0xFF)) {
    lcd.print("Attention.: MFRC522");
    lcd.setCursor(0, 1);
    lcd.print("Ctroller connection");
    delay(4000);
    while (true); // do not go further
  }
}
                                                                 

//////////////////////////////////////// Read an ID from EEPROM //////////////////////////////
void readID( uint8_t number ) {
  uint8_t start = (number * 4 ) + 2;    // Figure out starting position
  for ( uint8_t i = 0; i < 4; i++ ) {     // Loop 4 times to get the 4 Bytes
    storedCard[i] = EEPROM.read(start + i);   // Assign values read from EEPROM to array
  }
}

///////////////////////////////////////// Add ID to EEPROM   ///////////////////////////////////
void writeID( byte a[] ) {
  if ( !findID( a ) ) {     // Before we write to the EEPROM, check to see if we have seen this card before!
    uint8_t num = EEPROM.read(0);     // Get the numer of used spaces, position 0 stores the number of ID cards
    uint8_t start = ( num * 4 ) + 6;  // Figure out where the next slot starts
    num++;                // Increment the counter by one
    EEPROM.write( 0, num );     // Write the new count to the counter
    for ( uint8_t j = 0; j < 4; j++ ) {   // Loop 4 times
      EEPROM.write( start + j, a[j] );  // Write the array values to EEPROM in the right position
    }
    successWrite();
  }
  else {
    failedWrite();
  }
}

///////////////////////////////////////// Remove ID from EEPROM   ///////////////////////////////////
void deleteID( byte a[] ) {
  if ( !findID( a ) ) {     // Before we delete from the EEPROM, check to see if we have this card!
    failedWrite();      // If not
    delay (2000);
  }
  else {
    uint8_t num = EEPROM.read(0);   // Get the numer of used spaces, position 0 stores the number of ID cards
    uint8_t slot;       // Figure out the slot number of the card
    uint8_t start;      // = ( num * 4 ) + 6; // Figure out where the next slot starts
    uint8_t looping;    // The number of times the loop repeats
    uint8_t j;
    uint8_t count = EEPROM.read(0); // Read the first Byte of EEPROM that stores number of cards
    slot = findIDSLOT( a );   // Figure out the slot number of the card to delete
    start = (slot * 4) + 2;
    looping = ((num - slot) * 4);
    num--;      // Decrement the counter by one
    EEPROM.write( 0, num );   // Write the new count to the counter
    for ( j = 0; j < looping; j++ ) {         // Loop the card shift times
      EEPROM.write( start + j, EEPROM.read(start + 4 + j));   // Shift the array values to 4 places earlier in the EEPROM
    }
    for ( uint8_t k = 0; k < 4; k++ ) {         // Shifting loop
      EEPROM.write( start + j + k, 0);
    }
    successDelete();
    delay (2000);
  }
}

///////////////////////////////////////// Check Bytes   ///////////////////////////////////
bool checkTwo ( byte a[], byte b[] ) {   
  for ( uint8_t k = 0; k < 4; k++ ) {   // Loop 4 times
    if ( a[k] != b[k] ) {     // IF a != b then false, because: one fails, all fail
       return false;
    }
  }
  return true;  
}

///////////////////////////////////////// Find Slot   ///////////////////////////////////
uint8_t findIDSLOT( byte find[] ) {
  uint8_t count = EEPROM.read(0);       // Read the first Byte of EEPROM that
  for ( uint8_t i = 1; i <= count; i++ ) {    // Loop once for each EEPROM entry
    readID(i);                // Read an ID from EEPROM, it is stored in storedCard[4]
    if ( checkTwo( find, storedCard ) ) {   // Check to see if the storedCard read from EEPROM
      // is the same as the find[] ID card passed
      return i;         // The slot number of the card
    }
  }
}

///////////////////////////////////////// Find ID From EEPROM   ///////////////////////////////////
bool findID( byte find[] ) {
  uint8_t count = EEPROM.read(0);     // Read the first Byte of EEPROM that
  for ( uint8_t i = 1; i < count; i++ ) {    // Loop once for each EEPROM entry
    readID(i);          // Read an ID from EEPROM, it is stored in storedCard[4]
    if ( checkTwo( find, storedCard ) ) {   // Check to see if the storedCard read from EEPROM
      return true;
    }
    else {    // If not, return false
    }
  }
  return false;
}

///////////////////////////////////////// Write Success to EEPROM   ///////////////////////////////////
void successWrite() {
    lcd.clear();
    lcd.print("Succes: Ce Badge a");
    lcd.setCursor(0, 1);
    lcd.print( "ete ajoute dans la");
    lcd.setCursor(0, 2);
    lcd.print( "memeoire EEPROM");
    delay(3000);
}

///////////////////////////////////////// Write Failed to EEPROM   ///////////////////////////////////
void failedWrite() {
    lcd.clear();
    lcd.print("Erreur! Mauvaise ID");
    lcd.setCursor(0, 1);
    lcd.print( "ou Probleme memeoire");
    lcd.setCursor(0, 2);
    lcd.print( "ou EEPROM pleine !");
    delay(4000);
}

///////////////////////////////////////// Success Remove UID From EEPROM  ///////////////////////////////////
void successDelete() {
    lcd.clear();
    lcd.print("Succes:  Ce Badge a");
    lcd.setCursor(0, 1);
    lcd.print( "ete efface de la");
    lcd.setCursor(0, 2);
    lcd.print( "memeoire EEPROM");
    delay(3000); 
}

////////////////////// Check readCard IF is masterCard   ///////////////////////////////////
// Check to see if the ID passed is the master programing card
bool isMaster( byte test[] ) {
  return checkTwo(test, masterCard);
}

bool monitorWipeButton(uint32_t interval) {
  uint32_t now = (uint32_t)millis();
  while ((uint32_t)millis() - now < interval)  {
    // check on every half a second
    if (((uint32_t)millis() % 500) == 0) {
          lcd.backlight();
      if (digitalRead(wipeB) != LOW)
        return false;
    }
  }
  return true;
}

